/* Copyright 1998-2014 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */

#include "qubfast.h"
#include "qub_baseline_nodes.h"

#include <map>
using namespace std;
using std::map;

#ifndef NULL
#define NULL 0
#endif
 
// extend public types with private maps and iterators:
class QUB_BaselineSeg_NodeX : public QUB_BaselineSeg_Node {
public:
	std::map<int, QUB_BaselineSeg_NodeX*>::iterator iter;
};

class QUB_BaselineSegX : public QUB_BaselineSeg {
public:
	std::map<int, QUB_BaselineSeg_NodeX*> *nodes;
};


// exported

extern "C"     QUBFAST_API QUB_BaselineSeg *       QUB_BaselineSeg_Create( int first, int last )
{
	QUB_BaselineSegX *seg = new QUB_BaselineSegX;
	seg->first = first;
	seg->last = last;
	seg->nodes = new map<int, QUB_BaselineSeg_NodeX*>;

	return (QUB_BaselineSeg *) seg;
}

extern "C"     QUBFAST_API void                    QUB_BaselineSeg_Free( QUB_BaselineSeg *seg_ )
{
	QUB_BaselineSegX *seg = (QUB_BaselineSegX *) seg_;
	for ( map<int, QUB_BaselineSeg_NodeX*>::iterator ni = seg->nodes->begin(); ni != seg->nodes->end(); ++ni )
		delete ni->second;
	delete seg->nodes;
	delete seg;
}

extern "C"     QUBFAST_API int                     QUB_BaselineSeg_GetNodeCount( QUB_BaselineSeg *seg_ )
{
	QUB_BaselineSegX *seg = (QUB_BaselineSegX *) seg_;
	return (int) seg->nodes->size();
}

extern "C"     QUBFAST_API QUB_BaselineSeg_Node *  QUB_BaselineSeg_GetFirstNode( QUB_BaselineSeg *seg_ )
{
	QUB_BaselineSegX *seg = (QUB_BaselineSegX *) seg_;
	if ( seg->nodes->size() )
		return (QUB_BaselineSeg_Node *) seg->nodes->begin()->second;
	return NULL;
}

extern "C"     QUBFAST_API QUB_BaselineSeg_Node *  QUB_BaselineSeg_GetLastNode( QUB_BaselineSeg *seg_ )
{
	QUB_BaselineSegX *seg = (QUB_BaselineSegX *) seg_;
	if ( seg->nodes->size() ) {
		map<int, QUB_BaselineSeg_NodeX*>::iterator ni = seg->nodes->end();
		--ni;
		return (QUB_BaselineSeg_Node *) ni->second;
	}
	return NULL;
}

extern "C"     QUBFAST_API QUB_BaselineSeg_Node *  QUB_BaselineSeg_GetNodeBefore( QUB_BaselineSeg *seg_, QUB_BaselineSeg_Node *node_ )
{
	QUB_BaselineSegX *seg = (QUB_BaselineSegX *) seg_;
	QUB_BaselineSeg_NodeX *node = (QUB_BaselineSeg_NodeX *) node_;
	if ( ! node )
		return QUB_BaselineSeg_GetLastNode( seg_ );

	if ( node->point == seg->first )
		return NULL;

	map<int, QUB_BaselineSeg_NodeX*>::iterator ni = node->iter;
	if ( ni == seg->nodes->begin() )
		return NULL;

	--ni;
	return (QUB_BaselineSeg_Node *) ni->second;
}

extern "C"     QUBFAST_API QUB_BaselineSeg_Node *  QUB_BaselineSeg_GetNodeAfter( QUB_BaselineSeg *seg_, QUB_BaselineSeg_Node *node_ )
{
	QUB_BaselineSegX *seg = (QUB_BaselineSegX *) seg_;
	QUB_BaselineSeg_NodeX *node = (QUB_BaselineSeg_NodeX *) node_;
	if ( ! node )
		return QUB_BaselineSeg_GetFirstNode( seg_ );

	if ( node->point == seg->last )
		return NULL;

	map<int, QUB_BaselineSeg_NodeX*>::iterator ni = node->iter;
	++ni;
	if ( ni == seg->nodes->end() )
		return NULL;

	return (QUB_BaselineSeg_Node *) ni->second;
}


extern "C"     QUBFAST_API float QUB_BaselineSeg_InterpolateValueAt( QUB_BaselineSeg *seg_, int point )
{
	QUB_BaselineSeg_NodeX *from, *to;
	if ( QUB_BaselineSeg_GetRangeAt( seg_, point, (QUB_BaselineSeg_Node**)&from, (QUB_BaselineSeg_Node**)&to ) )
		return 0;
	double slope = ((double)(to->value - from->value)) / (to->point - from->point);
	float value = (float) ( from->value + slope * (point - from->point) );
	return value;
}

extern "C"     QUBFAST_API int   QUB_BaselineSeg_GetRangeAt( QUB_BaselineSeg *seg_, int point, QUB_BaselineSeg_Node **from_, QUB_BaselineSeg_Node **to_ )
{
	QUB_BaselineSegX *seg = (QUB_BaselineSegX *) seg_;
	QUB_BaselineSeg_NodeX **from = (QUB_BaselineSeg_NodeX **) from_;
	QUB_BaselineSeg_NodeX **to = (QUB_BaselineSeg_NodeX **) to_;
	if ( ! seg->nodes->size() ) {
		*from = *to = NULL;
		return -1;
	}

	map<int, QUB_BaselineSeg_NodeX*>::iterator ni = seg->nodes->upper_bound( point );
	if ( ni == seg->nodes->end() )
		--ni;
	if ( ni == seg->nodes->begin() )
		++ni;

	*to = ni->second;
	--ni;
	*from = ni->second;

	return 0;
}

extern "C"     QUBFAST_API int   QUB_BaselineSeg_GetRangeBefore( QUB_BaselineSeg *seg_, QUB_BaselineSeg_Node **from_, QUB_BaselineSeg_Node **to_ )
{
	QUB_BaselineSegX *seg = (QUB_BaselineSegX *) seg_;
	QUB_BaselineSeg_NodeX **from = (QUB_BaselineSeg_NodeX **) from_;
	QUB_BaselineSeg_NodeX **to = (QUB_BaselineSeg_NodeX **) to_;
	if ( ! from )
		return -1;
		
	map<int, QUB_BaselineSeg_NodeX*>::iterator ni = (*from)->iter;
	if ( ni == seg->nodes->begin() )
		return -1;

	*to = *from;
	--ni;
	*from = ni->second;

	return 0;
}

extern "C"     QUBFAST_API int   QUB_BaselineSeg_GetRangeAfter( QUB_BaselineSeg *seg_, QUB_BaselineSeg_Node **from_, QUB_BaselineSeg_Node **to_ )
{
	QUB_BaselineSegX *seg = (QUB_BaselineSegX *) seg_;
	QUB_BaselineSeg_NodeX **from = (QUB_BaselineSeg_NodeX **) from_;
	QUB_BaselineSeg_NodeX **to = (QUB_BaselineSeg_NodeX **) to_;
	if ( ! to )
		return -1;

	map<int, QUB_BaselineSeg_NodeX*>::iterator ni = (*to)->iter;
	++ni;
	if ( ni == seg->nodes->end() )
		return -1;

	*from = *to;
	*to = ni->second;

	return 0;
}


extern "C"     QUBFAST_API QUB_BaselineSeg_Node *  QUB_BaselineSeg_AddNode( QUB_BaselineSeg *seg_, int point, float value )
{
	QUB_BaselineSegX *seg = (QUB_BaselineSegX *) seg_;
	QUB_BaselineSeg_ClearNodes( seg, point, point, 0 );

	QUB_BaselineSeg_NodeX *node = new QUB_BaselineSeg_NodeX;
	node->point = point;
	node->value = value;
	(*seg->nodes)[point] = node;
	node->iter = seg->nodes->find(point);

	if ( seg->nodes->find( seg->first ) == seg->nodes->end() )
		QUB_BaselineSeg_AddNode( seg_, seg->first, value );
	
	if ( seg->nodes->find( seg->last ) == seg->nodes->end() )
		QUB_BaselineSeg_AddNode( seg_, seg->last, value );
	else if ( QUB_BaselineSeg_GetNodeAfter(seg_, node) == QUB_BaselineSeg_GetLastNode(seg_) )
		QUB_BaselineSeg_GetLastNode(seg_)->value = value;

	return (QUB_BaselineSeg_Node *) node;
}

extern "C"     QUBFAST_API QUB_BaselineSeg_Node *  QUB_BaselineSeg_AddNodes( QUB_BaselineSeg *seg_, int count, int *points, float *values )
{
	QUB_BaselineSegX *seg = (QUB_BaselineSegX *) seg_;
	// keep it in-bounds
	while ( count && (points[0] < seg->first) ) {
		--count;
		++points;
		++values;
	}
	while ( count && (points[count-1] > seg->last) ) {
		--count;
	}

	if ( count <= 0 )
		return NULL;

	QUB_BaselineSeg_ClearNodes( seg_, points[0], points[ count-1 ], 0 );

	QUB_BaselineSeg_NodeX *firstNode = NULL, *node;

	for ( int i=0; i<count; ++i ) {
		node = new QUB_BaselineSeg_NodeX;
		node->point = points[i];
		node->value = values[i];
		(*seg->nodes)[node->point] = node;
		node->iter = seg->nodes->find(node->point);

		if ( i == 0 )
			firstNode = node;
	}

	if ( seg->nodes->find( seg->first ) == seg->nodes->end() )
		QUB_BaselineSeg_AddNode( seg_, seg->first, values[0] );
	if ( seg->nodes->find( seg->last ) == seg->nodes->end() )
		QUB_BaselineSeg_AddNode( seg_, seg->last, values[ count-1 ] );
	else if ( QUB_BaselineSeg_GetNodeAfter(seg, node) == QUB_BaselineSeg_GetLastNode(seg_) )
		QUB_BaselineSeg_GetLastNode(seg_)->value = values[ count-1 ];

	return firstNode;
}

extern "C"     QUBFAST_API void                    QUB_BaselineSeg_ClearNodes( QUB_BaselineSeg *seg_, int first, int last, int repairFirstLast)
{
	QUB_BaselineSegX *seg = (QUB_BaselineSegX *) seg_;
	map<int, QUB_BaselineSeg_NodeX*>::iterator ni = seg->nodes->lower_bound( first );
	if ( ni == seg->nodes->end() )
		return;

	QUB_BaselineSeg_NodeX *node = ni->second;
	QUB_BaselineSeg_NodeX *next = (QUB_BaselineSeg_NodeX *) QUB_BaselineSeg_GetNodeAfter( seg_, node );
	while ( next && (node->point <= last) ) {
		seg->nodes->erase( node->iter );
		delete node;
		node = next;
		next = (QUB_BaselineSeg_NodeX *) QUB_BaselineSeg_GetNodeAfter( seg_, node );
	}
	if ( node->point <= last ) {
		seg->nodes->erase( node->iter );
		delete node;
	}

	if ( repairFirstLast && seg->nodes->size() ) {
		if ( seg->nodes->find( seg->first ) == seg->nodes->end() )
			QUB_BaselineSeg_AddNode( seg_, seg->first, seg->nodes->begin()->second->value );
		if ( seg->nodes->find( seg->last ) == seg->nodes->end() ) {
			ni = seg->nodes->end();
			--ni;
			QUB_BaselineSeg_AddNode( seg_, seg->last, ni->second->value );
		}
	}
}

